// Copyright (C) 2016-Present Pivotal Software, Inc. All rights reserved.
// This program and the accompanying materials are made available under the terms of the under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// http://www.apache.org/licenses/LICENSE-2.0
// Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the specific language governing permissions and limitations under the License.

package upgrader_test

import (
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf/brokerapi"
	"github.com/pivotal-cf/on-demand-service-broker/broker"
	"github.com/pivotal-cf/on-demand-service-broker/mockbroker"
	"github.com/pivotal-cf/on-demand-service-broker/mockhttp"
	"github.com/pivotal-cf/on-demand-service-broker/upgrader"
	"github.com/pivotal-cf/on-demand-service-broker/upgrader/broker_response"
)

var _ = Describe("Broker Services HTTP Client", func() {
	const (
		brokerUsername      = "username"
		brokerPassword      = "password"
		serviceInstanceGUID = "my-service-instance"
		invalidURL          = "Q#$%#$%^&&*$%^#$FGRTYW${T:WED:AWSD)E@#PE{:QS:{QLWD"
	)

	var (
		odb    *mockhttp.Server
		client upgrader.BrokerServicesHTTPClient
	)

	BeforeEach(func() {
		odb = mockbroker.New()
		odb.ExpectedBasicAuth(brokerUsername, brokerPassword)
		client = upgrader.NewBrokerServicesHTTPClient(brokerUsername, brokerPassword, odb.URL)
	})

	AfterEach(func() {
		odb.VerifyMocks()
		odb.Close()
	})

	Describe("Instances", func() {
		It("returns a list of instances", func() {
			odb.VerifyAndMock(
				mockbroker.ListInstances().RespondsWith(`[{"instance_id": "foo"}, {"instance_id": "bar"}]`),
			)

			instances, err := client.Instances()

			Expect(err).NotTo(HaveOccurred())
			Expect(instances).To(ConsistOf("foo", "bar"))
		})

		Context("when the url is invalid", func() {
			It("returns an error", func() {
				client := upgrader.NewBrokerServicesHTTPClient(brokerUsername, brokerPassword, invalidURL)

				_, err := client.Instances()

				Expect(err).To(HaveOccurred())
			})
		})

		Context("when the request fails", func() {
			It("returns an error", func() {
				odb.Close()

				_, err := client.Instances()

				Expect(err).To(HaveOccurred())
			})
		})

		Context("when the broker response is unrecognised", func() {
			It("returns an error", func() {
				odb.VerifyAndMock(
					mockbroker.ListInstances().RespondsWith(`{"not": "a list"}`),
				)

				_, err := client.Instances()

				Expect(err).To(HaveOccurred())
			})
		})
	})

	Describe("UpgradeInstance", func() {
		It("returns an upgrade operation", func() {
			odb.VerifyAndMock(
				mockbroker.UpgradeInstance(serviceInstanceGUID).NotFound(),
			)

			upgradeOperation, err := client.UpgradeInstance(serviceInstanceGUID)

			Expect(err).NotTo(HaveOccurred())
			Expect(upgradeOperation.Type).To(Equal(broker_response.ResultNotFound))
		})

		Context("when the url is invalid", func() {
			It("returns an error", func() {
				client := upgrader.NewBrokerServicesHTTPClient(brokerUsername, brokerPassword, invalidURL)

				_, err := client.UpgradeInstance(serviceInstanceGUID)

				Expect(err).To(HaveOccurred())
			})
		})

		Context("when the request fails", func() {
			It("returns an error", func() {
				odb.Close()

				_, err := client.UpgradeInstance(serviceInstanceGUID)

				Expect(err).To(HaveOccurred())
			})
		})

		Context("when the broker responds with an error", func() {
			It("returns an error", func() {
				odb.VerifyAndMock(
					mockbroker.UpgradeInstance(serviceInstanceGUID).Fails("error upgrading instance"),
				)

				_, err := client.UpgradeInstance(serviceInstanceGUID)

				Expect(err).To(HaveOccurred())
			})
		})
	})

	Describe("LastOperation", func() {
		It("returns a last operation", func() {
			operationData := broker.OperationData{
				BoshTaskID:    1,
				BoshContextID: "context-id",
				OperationType: broker.OperationTypeUpgrade,
				PlanID:        "plan-id",
			}
			expectedOperationDataJSON := `{"BoshTaskID":1,"BoshContextID":"context-id","OperationType":"upgrade","PlanID":"plan-id"}`
			odb.VerifyAndMock(
				mockbroker.LastOperation(serviceInstanceGUID, expectedOperationDataJSON).
					RespondsWith(`{"state":"in progress","description":"upgrade in progress"}`),
			)

			lastOperation, err := client.LastOperation(serviceInstanceGUID, operationData)

			Expect(err).NotTo(HaveOccurred())
			Expect(lastOperation).To(Equal(
				brokerapi.LastOperation{State: brokerapi.InProgress, Description: "upgrade in progress"}),
			)
		})

		Context("when the url is invalid", func() {
			It("returns an error", func() {
				client := upgrader.NewBrokerServicesHTTPClient(brokerUsername, brokerPassword, invalidURL)

				_, err := client.LastOperation(serviceInstanceGUID, broker.OperationData{})

				Expect(err).To(HaveOccurred())
			})
		})

		Context("when the request fails", func() {
			It("returns an error", func() {
				odb.Close()

				_, err := client.LastOperation(serviceInstanceGUID, broker.OperationData{})

				Expect(err).To(HaveOccurred())
			})
		})

		Context("when the broker response is unrecognised", func() {
			It("returns an error", func() {
				expectedOperationDataJSON := `{"BoshTaskID":0,"OperationType":""}`
				odb.VerifyAndMock(
					mockbroker.LastOperation(serviceInstanceGUID, expectedOperationDataJSON).
						RespondsWith("invalid json"),
				)

				_, err := client.LastOperation(serviceInstanceGUID, broker.OperationData{})

				Expect(err).To(HaveOccurred())
			})
		})
	})
})
